    extrn _iprint   : near     ; number in stack
    extrn _bprint   : near     ; byte in stack
    extrn _getstr   : near     ; get string, no parameters, returns address in ax
    extrn _atoi     : near
    extrn _IsSegment: near

DisplayFlag macro NAME, POSITION
    local flagclear
    local flagset
    local printflag
    mov ax, 0900h
    mov dx, offset NAME; 
    int 21h
    mov ax, _flags
    and ax, POSITION
    jz flagclear
    flagset:
    mov al, 49
    jmp printflag
    flagclear:
    mov al, 48
    printflag:
    mov ah, 0EH
    int 10h
endm

DisplayARegister macro NAME, REGISTER
    mov ax, 0900h
    mov dx, offset NAME; 
    int 21h
    push REGISTER
    call _iprint
    add sp, 2
endm

PrintInteger macro REGVAL
    push REGVAL
    call _iprint
    add sp, 2
endm

PrintByte macro REGVAL
    push REGVAL
    call _bprint
    add sp, 2
endm

PrintString macro STRING
    push ax
    push dx
    mov ax, 0900h
    mov dx, offset STRING
    int 21h
    pop dx
    pop ax
endm

cr equ 0dh
lf equ 0ah

_stack segment stack 'stack'
    db 512 dup(0)
_stack ends

_text segment  public 'code'
    assume cs: _text, ds:_DATA, ss:_stack, es:_data

_isr proc far
    sti
    ;saving registers
    push ds
    push ax

    ;initialising ds
    mov ax, _data
    mov ds, ax
    
    mov _bx, bx
    mov _cx, cx
    mov _dx, dx
    mov _si, si
    mov _di, di
    mov _bp, bp
    mov _es, es
    mov _ss, ss

    ;calculating sp 
    mov ax,  sp
    add ax,  12D
    mov _sp, ax

    ;getting cs, ip, ds, ax from stack and printing them
    mov bp, sp
    mov ax, [bp + 6]; cs
    mov _cs, ax
    mov ax, [bp + 4]; ip
    mov _ip, ax
    mov ax, [bp + 2 ]; ds
    mov _ds, ax
    mov ax, [bp]; ds
    mov _ax, ax
    mov ax, [bp + 8]; cs
    mov _flags, ax

    ;;Checking for trace mode
    cmp trace, 0
    jne process

    ;;checking if breaking point has been reached
    mov ax, breakpoint
    cmp _ip, ax
    jnz retnear
    mov ax, breakpoint+2
    cmp _cs, ax
    jnz retnear
    mov trace, 1
    jmp process

    retnear:
    jmp restore_reg

    process:
    call DispInst
    call DispReg

    ;Waiting for response
    command:
    PrintString prompt
    mov ah, 00H
    int 16h
    mov ah, 14D
    int 10h
    cmp al, 't'
    je restore_reg
    cmp al, 'T'
    je restore_reg
    cmp al, 'g'
    je NoTrace
    cmp al, 'G'
    je NoTrace
    cmp al, 'r'
    je Registers
    cmp al, 'R'
    je Registers
    cmp al, 'd'
    je MemDump
    cmp al, 'D'
    je MemDump

    InvalidCommand:
    PrintString invalidcmd
    jmp command

    Registers:
    call DispReg
    jmp command

    MemDump:
        push _ds
        push 0000H
        call GetParams
        call Dump
        add sp, 4
        jmp command

    NoTrace:
        mov ax, 0
        push ax
        push ax
        call GetParams
        pop ax
        pop bx
        mov breakpoint,     ax
        mov breakpoint+2,   bx
        mov trace, 0
        PrintString newline
   restore_reg:
    mov bx, _bx
    mov cx, _cx
    mov dx, _dx
    mov si, _si
    mov di, _di
    mov bp, _bp
    mov es, _es
    
   return:
    pop ax
    pop ds
    iret
_isr endp

DispInst proc near

    PrintString Instruction

    mov ax, _cs
    mov es, ax
    mov bx, _ip
    mov cl, es:[bx]
    mov ch, 0
    PrintByte cx
    ret
DispInst endp

DispReg proc near

    push ax
    push dx
    DisplayARegister Reg_AX, _ax
    DisplayARegister Reg_BX, _bx
    DisplayARegister Reg_CX, _cx
    DisplayARegister Reg_DX, _dx
    DisplayARegister Reg_SP, _sp
    DisplayARegister Reg_BP, _bp
    DisplayARegister Reg_SI, _si
    DisplayARegister Reg_DI, _di
    DisplayARegister Reg_DS, _ds
    DisplayARegister Reg_ES, _es
    DisplayARegister Reg_SS, _ss
    DisplayARegister Reg_CS, _cs
    DisplayARegister Reg_IP, _ip
    DisplayFlag Reg_Carry 0001H
    DisplayFlag Reg_Parity 0004H
    DisplayFlag Reg_Aux 0010H
    DisplayFlag Reg_Zero 0040H
    DisplayFlag Reg_Sign 0080H
    DisplayFlag Reg_IE 0200H
    DisplayFlag Reg_Direction 0400H
    DisplayFlag Reg_Overflow 0800H
    pop dx
    pop ax

    ret

DispReg endp

_main   proc far

    mov ax, _data
    mov ds, ax

    ;Getting Command Line
    mov bx, 81H
    mov si, 1
    mov ch, 0
    mov cl, es: 80H
    cmp cl, 0
    jne getprgname
    PrintString Copyright
    PrintString PrgNotSpecified
    jmp back_to_dos

    getprgname:
        cmp byte ptr es:[bx][si], ' '
        je donegetting
        cmp byte ptr es:[bx][si], 0DH
        je donegetting
        mov al, byte ptr es: [bx][si]
        mov prognam-1 [si], al
        inc si
        loop getprgname
    donegetting:
    mov prognam-1 [si], 0

     ;resizing program
        mov ah, 4ah
        assume cs:zndseg
        mov bx, seg zndseg
        assume cs:_text
        mov cx, es
        sub bx, cx
        mov prgsize, bx
        int 21h
        jc err_resize

      ;getting isr 1
        mov ah, 35h 
        mov al, 1 ;;trace interrupt
        int 21h
        mov oldisr, bx
        mov oldisr+2, es

      ;setting isr 1
        mov ah, 25h
        mov al, 1 ;;trace interrupt
        mov dx, seg _isr
        mov ds, dx
        mov dx, offset _isr
        int 21h

    ;initialising data and extra segment registers
        mov ax, _data
        mov ds, ax
        mov es, ax
        PrintString Copyright

    ;;;loading child and tracing
        call _load

    ;resetting isr
        mov ah, 25h
        mov al, 1 ;;trace interrupt
        mov dx, oldisr + 2    ;segment
        mov dx, oldisr        ;offset
        int 21h
        jmp back_to_dos

    ;unable to resize program: fatal error
    err_resize:
        PrintString msg1; error resizeing program
        mov al, 01

    back_to_dos:
        mov ah, 4ch
        int 21h
    _main endp

_load proc near

    ;printing message
        PrintString msg3     ; executing program...

        ;Saving ES
        push es

        ;;loading child program
        mov ah, 4bh
        mov al, 01h
        lea bx, pararea
        lea dx, prognam
        int 21h
        jc err_load

        ;saving child's cs:ip in _text segment
        mov bx, word ptr csip
        mov cs:ip_cs, bx
        mov bx, word ptr csip+2
        mov cs:cs_cs, bx

        ; saving segments registers

        ;saving stack pointer
        mov ax, sp
        mov parent_sp, ax


        ;get psp
        mov ah, 51h
        int 21h

        ;setting return address
        mov es, bx
        mov word ptr es:0ah, offset normal_termination
        mov word ptr es:0ch, seg normal_termination


        ;setting up stack and data segments for child
        mov ax, sssp
        mov sp, ax
        mov ax, sssp+2
        mov ss, ax
        mov ds, bx

    ;**********passing control to child process***********
        mov bx, offset cs:csip_cs

        ;;setting trace flag on
        pushf
        mov bp, sp
        or [bp+0],0100h
        popf

        jmp dword ptr cs:[bx]

    err_load:
        PrintString msg2; error loading program
        mov al, 02
        pop es
        jmp ret_main
    
    normal_termination:

        ;restoring registers
        mov ax, _data
        mov ds, ax
        mov ax, parent_sp
        mov sp, ax
        pop es

        PrintString msg4; normal program termination

    ;_iprint calls
    ;    mov ax, sp
    ;    PrintInteger ax
    ;    mov ax, ds
    ;    PrintInteger ax
    ;    mov ax, es
    ;    PrintInteger ax
    ;    mov ax, ss
    ;    PrintInteger ax
        ;;;end call

    ret_main:
        ret
_load endp

Dump proc near
    push es
    push bp
    push ax
    push bx
    push cx
    push si
    push di

    PrintString newline
    mov bp, sp
    mov ax, [bp+18]
    mov es, ax
    mov bx, [bp+16]

    ;; Segment:Offset - es:bx

    mov di, 8
    loop_main:
        PrintString newline
        PrintInteger es
        PrintString colon    
        PrintInteger bx
        mov ax, 0E20H; print a space
        int 10h
        mov ah, 0
        mov si, 00H
        mov cx, 10H
        loop1:
            mov al, es:[bx][si]
            PrintByte ax
            add si, 1
        loop loop1
        PrintString hyphen
        mov si, 00H
        mov cx, 10H
        loop2:
            mov al, es:[bx][si]
            cmp al, 32
            jge printascii
            mov al, 46
            printascii:
            mov ah, 0EH;;ascii
            int 10h
            add si, 1
        loop loop2
        add bx, 16
        dec di
    jnz loop_main

    pop di
    pop si
    pop cx
    pop bx
    pop ax
    pop bp
    pop es
    ret
Dump endp

GetParams proc near
    ;receives the default segment:offset in stack
    ;returns after modifying them if necessary
    push bx
    push bp

    mov ax, 0E20H
    int 10H
    call _getstr    ; returns pointer to string in ax
    mov bx, ax
    cmp byte ptr [bx], 0
    je noparams
    params:
        inc bx
        push bx
        call _IsSegment
        jne segparams

        add sp, 2
        push bx
        call _atoi
        pop bx

        mov bp, sp
        mov word ptr [bp+6], ax
        ;;Checking if offset is specified
        skiploop1: ; Skipping blanks
            cmp byte ptr [bx], ' '
            jne afterskipping1
            inc bx
            jmp skiploop1
        afterskipping1:
        cmp byte ptr [bx], ':'
        jne noparams
        inc bx
        push bx
        call _atoi
        pop bx
        mov bp, sp
        mov bx, word ptr [bp+6]
        mov word ptr [bp+8], bx
        mov word ptr [bp+6], ax
        jmp noparams

        segparams:
        pop bx
        cmp ax, 1   ;ds
        jz DSSeg
        cmp ax, 2   ;es
        jz ESSeg
        cmp ax, 3   ;cs
        jz CSSeg
        cmp ax, 4   ;ss
        jz SSSeg
        
        SSSeg:
        mov ax, _ss
        jmp moveseg
        DSSeg:
        mov ax, _ds
        jmp moveseg
        ESSeg:
        mov ax, _es
        jmp moveseg
        CSSeg:
        mov ax, _cs

        moveSeg:
        mov bp, sp
        mov word ptr [bp+8], ax ; changing segment
        ;;Checking if offset is specified
        skiploop: ; Skipping blanks
            cmp byte ptr [bx], ' '
            jne afterskipping
            inc bx
            jmp skiploop
        afterskipping:
        cmp byte ptr [bx], ':'
        jne noparams
        inc bx
        push bx
        call _atoi
        pop bx
        mov bp, sp
        mov word ptr [bp+6], ax
    noparams:
    pop bp
    pop bx
    ret
GetParams endp

    csip_cs label dword
        ip_cs dw 0
        cs_cs dw 0

_text ends


_data segment para public 'data' 

    pararea    label byte
    dw 00h
    dw offset param
    dw _data
    dw offset fcb1
    dw _data
    dw offset fcb2
    dw _data
    sssp    dw 0
            dw 0
    csip    dd 0
    parent_sp dw 0
    dircom  db  0
    param   db 20 dup(' ')
    fcb1    db 0, 11 dup (32), 25 dup(0)
    fcb2    db 0, 11 dup (32), 25 dup(0)

    prognam db 30 dup(0)    ;'three.exe',0

    prgsize dw 0        ;holds the size of the program
    oldisr label word  ;old interrupt service routine
            dw 0
            dw 0

    ;Messages *******************
    Copyright   db "Debugger 2000"
                db cr, lf, 'Copyright (c) Manoj Paul Joseph 2000'
                db cr, lf, "www.FreeSourceCodes.org",cr, lf, '$'
    PrgNotSpecified db cr, lf, "Error: Program name not specified"
                db cr, lf, "Syntax: Load <Executable file>$"
    msg1    db cr, lf, "Error resizeing program",cr,lf,'$'
    msg2    db cr, lf, "Error loading program",cr,lf,'$'
    msg3    db cr, lf, "Executing program...",cr,lf,'$'
    msg4    db cr, lf, "Normal program termination",cr,lf,'$'
    newline    db cr, lf,'$'
    colon   db ':$'
    hyphen  db ' - $'
    Loading db cr, lf, "Loading Program : $"
    prompt  db cr, lf, "Command : $"
    invalidcmd  db cr, lf, "Invalid command! $"
    Instruction db cr, lf, cr, lf, "Instruction : $"

    Reg_DS  db cr, lf, 'DS: $'
    Reg_ES  db ' ES: $'
    Reg_SS  db ' SS: $'
    Reg_CS  db ' CS: $'
    Reg_IP  db ' IP: $'
    Reg_Carry       db cr, lf, 'Carry:$'
    Reg_Parity      db '  Parity:$'
    Reg_Aux         db '  AuxCarry:$'
    Reg_Zero        db '  Zero:$'
    Reg_Sign        db '  Sign:$'
    Reg_IE          db '  Inter:$'
    Reg_Direction   db '  Direction:$'
    Reg_Overflow    db '  Overflow:$'
    Reg_AX  db cr, lf,'AX: $'
    Reg_BX  db ' BX: $'
    Reg_CX  db ' CX: $'
    Reg_DX  db ' DX: $'
    Reg_SP  db ' SP: $'
    Reg_BP  db ' BP: $'
    Reg_SI  db ' SI: $'
    Reg_DI  db ' DI: $'

    trace       db 1 
    breakpoint  dw 0,0

    _ax dw 0
    _bx dw 0
    _cx dw 0
    _dx dw 0
    _si dw 0
    _di dw 0
    _sp dw 0
    _bp dw 0
    _cs dw 0
    _ds dw 0
    _es dw 0
    _ss dw 0
    _ip dw 0
    _flags dw 0

_data ends

zndseg  segment
    ;dummy segment
zndseg ends

end _main
